/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    Author Quan Sun (quan.sun.nz@gmail.com)
 *    Copyright (C) 2013 University of Waikato, Hamilton, New Zealand
 *
 */

package artforests.core;

import java.util.Random;
import org.apache.commons.math.stat.descriptive.DescriptiveStatistics;
import weka.core.Instances;

public abstract class AbstractRanker implements Ranker, java.io.Serializable{

    protected boolean m_Debug = true;

    public void setDebug(boolean d) {
        m_Debug = d;
    }

    public static double[] getCenterRanking(Instances data, String mode) throws Exception {
        if (mode.equals("AVG")) {
            return getAvgRanking(data);
        }
        if (mode.equals("MED")) {
            return getMedianRanking(data);
        }
        return getEmpricalCenterRanking(data);
    }

    public static double[] getEmpricalCenterRanking(Instances data) throws Exception {
        if (data.numInstances() == 0) {
            throw new Exception("getEmpricalCenterRanking: Dataset is empty");
        }

        double maxDis = 2.0;
        double avgDis[] = new double[data.numInstances()];
        for (int i = 0; i < data.numInstances(); i++) {
            MultiTargetInstance metaInst_i = (MultiTargetInstance) data.instance(i);
            for (int j = 0; j < data.numInstances(); j++) {
                if (i != j) {
                    MultiTargetInstance metaInst_j = (MultiTargetInstance) data.instance(j);
                    double score = EvalTool.computeScore(metaInst_i.getRanking(), metaInst_j.getRanking());
                    if (score >= 0) {
                        avgDis[i] += maxDis - (1 + score);
                    } else {
                        avgDis[i] += (1 + -1.0 * score);
                    }
                }
            }
            avgDis[i] /= (data.numInstances() - 1 + 0.000001);
        }

        int minDisIndex = weka.core.Utils.minIndex(avgDis);
        double[] trueRanking = Tools.doubleArrayToRanking(((MultiTargetInstance) data.instance(minDisIndex)).getRanking(), EvalTool.RANKING_ALGO);
        return trueRanking;
    }

    public static double[] getMedianRanking(Instances data) throws Exception {

        if (data.numInstances() == 0) {
            throw new Exception("getMedianRanking: Dataset is empty");
        }

        MultiTargetInstance metaInstx = (MultiTargetInstance) data.instance(0);
        double avgRank[] = new double[metaInstx.getRanking().length];
        DescriptiveStatistics[] stats = new DescriptiveStatistics[metaInstx.getRanking().length];
        for (int j = 0; j < avgRank.length; j++) {
            stats[j] = new DescriptiveStatistics();
        }
        for (int i = 0; i < data.numInstances(); i++) {
            MultiTargetInstance metaInst = (MultiTargetInstance) data.instance(i);
            for (int j = 0; j < avgRank.length; j++) {
                if (Double.isNaN(metaInst.getRanking()[j]) != true) {
                    //avgRank[j] += metaInst.getRanking()[j] * metaInst.weight();
                    stats[j].addValue(metaInst.getRanking()[j]);
                }
            }
        }
        for (int j = 0; j < avgRank.length; j++) {

            double median = stats[j].getPercentile(50);
            avgRank[j] = median;
        }
        double[] trueRanking = Tools.doubleArrayToRanking(avgRank, EvalTool.RANKING_ALGO);
        return trueRanking;
    }

    public static double[] getAvgRankingBagging(Instances data) throws Exception {

        if (data.numInstances() == 0) {
            throw new Exception("getAvgRank: Dataset is empty");
        }
        MultiTargetInstance metaInstx = (MultiTargetInstance) data.instance(0);
        double avgRanks[] = new double[metaInstx.getRanking().length];
        Random r = new Random(1);
        int baggingItr = 30;
        for (int bst = 0; bst < baggingItr; bst++) {
            for (int i = 0; i < data.numInstances(); i++) {
                MultiTargetInstance metaInst = (MultiTargetInstance) data.instance(r.nextInt(data.numInstances()));
                for (int j = 0; j < avgRanks.length; j++) {
                    avgRanks[j] += metaInst.getRanking()[j] * metaInst.weight();
                }
            }
        }
        for (int j = 0; j < avgRanks.length; j++) {
            avgRanks[j] /= (baggingItr * data.numInstances());
        }
        double[] trueRanking = Tools.doubleArrayToRanking(avgRanks, EvalTool.RANKING_ALGO);
        return trueRanking;
    }

    public static double[] getAvgRanking(Instances data) throws Exception {        
        if (data.numInstances() == 0) {
            throw new Exception("getAvgRank: Dataset is empty");
        }
        MultiTargetInstance metaInstx = (MultiTargetInstance) data.instance(0);
        double avgRanks[] = new double[metaInstx.getRanking().length];
        for (int i = 0; i < data.numInstances(); i++) {
            MultiTargetInstance metaInst = (MultiTargetInstance) data.instance(i);
            for (int j = 0; j < avgRanks.length; j++) {
                if (Double.isNaN(metaInst.getRanking()[j]) != true) {
                    avgRanks[j] += metaInst.getRanking()[j] * metaInst.weight();
                }
            }
        }
        double sumWeights = 0;
        for (int j = 0; j < data.numInstances(); j++) {
            MultiTargetInstance metaInst = (MultiTargetInstance) data.instance(j);
            sumWeights += metaInst.weight();
        }
        for (int j = 0; j < avgRanks.length; j++) {
            avgRanks[j] /= (1.0 * sumWeights);
        }
        double[] trueRanking = Tools.doubleArrayToRanking(avgRanks, EvalTool.RANKING_ALGO);
        return trueRanking;
    }

    public static double[] getAvgRankValues(Instances data) throws Exception {

        if (data.numInstances() == 0) {
            throw new Exception("getAvgRank: Dataset is empty");
        }
        MultiTargetInstance metaInstx = (MultiTargetInstance) data.instance(0);
        double avgRank[] = new double[metaInstx.getRanking().length];
        for (int i = 0; i < data.numInstances(); i++) {
            MultiTargetInstance metaInst = (MultiTargetInstance) data.instance(i);
            for (int j = 0; j < avgRank.length; j++) {
                if (Double.isNaN(metaInst.getRanking()[j]) != true) {
                    avgRank[j] += metaInst.getRanking()[j] * metaInst.weight();
                }
            }
        }
        double sumWeights = 0;
        for (int j = 0; j < data.numInstances(); j++) {
            MultiTargetInstance metaInst = (MultiTargetInstance) data.instance(j);
            sumWeights += metaInst.weight();
        }

        for (int j = 0; j < avgRank.length; j++) {
            avgRank[j] /= (1.0 * sumWeights);
        }
        return avgRank;
    }
}

